####################################################################
#                                                                  #
#    ensemble_md,                                                  #
#    a python package for running GROMACS simulation ensembles     #
#                                                                  #
#    Written by Wei-Tse Hsu <wehs7661@colorado.edu>                #
#    Copyright (c) 2022 University of Colorado Boulder             #
#                                                                  #
####################################################################
"""
Unit tests for the module gmx_parser.py. Note that most part of the functions
here have been automatically tested when testing some of the functions in ensemble_md,
so here we are just testing the untested part.
"""
import os
import pytest
import numpy as np
from ensemble_md.utils import gmx_parser
from ensemble_md.utils.exceptions import ParseError

current_path = os.path.dirname(os.path.abspath(__file__))
input_path = os.path.join(current_path, "data")


def test_parse_log():
    """
      - Case 1: The weights have never been equilibrated.
      - Case 2: The weights were equilibrated during the simulation.
      - Case 3: The weights were fixed in the simulation.
    """
    # Case 1: weight-updating simulation
    weights_0, counts_0, wl_delta_0, equil_time_0 = gmx_parser.parse_log(os.path.join(input_path, 'log/EXE_0.log'))
    assert len(weights_0) == 5
    assert weights_0[0] == [0.0, 3.83101, 4.95736, 5.63808, 6.0722, 6.13408]
    assert weights_0[-1] == [0.00000, 1.03101, 2.55736, 3.63808, 4.47220, 6.13408]
    assert counts_0 == [4, 11, 9, 9, 11, 6]
    assert wl_delta_0 == 0.4
    assert np.isclose(equil_time_0, -1)

    # Case 2-1: equil_time < 1000 ps
    # As a note, here we used a very lenient weight-equil-wl-delta as 0.3, with nst_sim = 2000.
    # The log file was from sim_0/iteration_1.
    weights_1, counts_1, wl_delta_1, equil_time_1 = gmx_parser.parse_log(os.path.join(input_path, 'log/case2_1.log'))
    assert len(weights_1) == 6
    assert weights_1[0] == [0.0, 2.68453, 4.13258, 4.3248, 4.4222, 6.52607]
    assert weights_1[-1] == [0.00000, 1.40453, 2.85258, 2.72480, 3.46220, 5.88607]
    assert counts_1 == [11, 13, 15, 4, 4, 50]
    assert wl_delta_1 is None
    assert np.isclose(equil_time_1, 6.06)

    # Case 2-2: equil_time > 1000 ps (just another case, nothing special here ...)
    # As a note, here we used weight-equil-wl-delta as 0.3, with nst_sim = 2000,  n_iterations = 1000, n_sim = 2,
    # s = 4, wl-scale = 0.9, and wl-ratio = 0.9. The log file was from sim_0/iteration_475 (t=1.9 ns).
    weights_2, counts_2, wl_delta_2, equil_time_2 = gmx_parser.parse_log(os.path.join(input_path, 'log/case2_2.log'))
    assert len(weights_2) == 19
    assert weights_2[0] == [0.00000, 1.60863, 2.47927, 3.15184, 3.47507]
    assert weights_2[-1] == [0.00000, 1.60419, 2.47705, 3.14963, 3.47840]
    assert counts_2 == [2, 3, 1, 2, 1]
    assert wl_delta_2 is None
    assert equil_time_2 == 1903.82

    # Case 3: fixed-weight simulation
    weights_3, counts_3, wl_delta_3, equil_time_3 = gmx_parser.parse_log(os.path.join(input_path, 'log/case3.log'))
    assert len(weights_3) == 1
    assert weights_3[-1] == [0.00000, 1.55165, 2.55043, 3.15034, 3.26889, 4.37831, 5.28574, 3.29638, 2.22527]
    assert counts_3 == [12, 9, 10, 9, 4, 5, 1, 0, 0]
    assert wl_delta_3 is None
    assert equil_time_3 == 0


def test_filename():
    MDP = gmx_parser.MDP()
    with pytest.raises(ValueError, match="A file name is required because no default file name was defined."):
        MDP.filename()

    MDP._filename = 'test'
    assert MDP.filename() == 'test'


class Test_MDP:
    def test__eq__(self):
        mdp_1 = gmx_parser.MDP("ensemble_md/tests/data/expanded.mdp")
        mdp_2 = gmx_parser.MDP("ensemble_md/tests/data/expanded.mdp")
        assert mdp_1 == mdp_2

    def test_read(self):
        mdp = gmx_parser.MDP()
        f = open("fake.mdp", "a")
        f.write("TEST")
        f.close()
        with pytest.raises(ParseError, match="'fake.mdp': unknown line in mdp file, 'TEST'"):
            mdp.read('fake.mdp')
        os.remove('fake.mdp')

    def test_write(self):
        mdp = gmx_parser.MDP("ensemble_md/tests/data/expanded.mdp")
        mdp.write('test_1.mdp', skipempty=False)
        mdp.write('test_2.mdp', skipempty=True)
        os.remove('test_1.mdp')
        os.remove('test_2.mdp')



####################################################################
#                                                                  #
#    ensemble_md,                                                  #
#    a python package for running GROMACS simulation ensembles     #
#                                                                  #
#    Written by Wei-Tse Hsu <wehs7661@colorado.edu>                #
#    Copyright (c) 2022 University of Colorado Boulder             #
#                                                                  #
####################################################################
"""
The :obj:`.analyze_free_energy` module provides functions for performing free energy
calculations for REXEE simulations.
"""
import alchemlyb
import numpy as np
import matplotlib.pyplot as plt
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

from pymbar.timeseries import detect_equilibration, subsample_correlated_data  # noqa: E402
from alchemlyb.estimators import TI, BAR, MBAR  # noqa: E402
from alchemlyb.parsing.gmx import extract_dHdl, extract_u_nk  # noqa: E402
from alchemlyb.preprocessing import subsampling  # noqa: E402
from ensemble_md.utils import utils  # noqa: E402
from ensemble_md.utils.exceptions import ParameterError  # noqa: E402


def preprocess_data(files_list, temp, data_type, spacing=1, t=None, g=None):
    """
    This function preprocesses :math:`u_{nk}`/:math:`dH/dλ` data for all replicas in an REXEE simulation.
    For each replica, it reads in :math:`u_{nk}`/:math:`dH/dλ` data from all iterations, concatenate
    them, remove the equilibrium region and and decorrelate the concatenated data. Notably,
    the data preprocessing protocol is basically the same as the one adopted in
    :code:`alchemlyb.subsampling.equilibrium_detection`.

    Parameters
    ----------
    files_list : list
        A list of lists of naturally sorted dhdl file names from all iterations for different replicas.
        :code:`files[i]` should be the list of dhdl file names from all iterations of replica :code:`i`.
    temp : float
        The simulation temperature in Kelvin. We assume all replicas were performed at the same temperature.
    data_type : str
        The type of energy data to be procssed. Should be either :code:`'u_nk'` or :code:`'dhdl'`.
    spacing : int
        The spacing (number of data points) to consider when subsampling the data, which is assumed to
        be the same for all replicas.
    t : int
        The user-specified index that indicates the start of equilibrated data. If this parameter is not specified,
        the function will estimate it using :code:`pymbar.timeseries.detect_equilibration`.
    g : int
        The user-specified index that indicates the start of equilibrated data. If this parameter is not specified,
        the function will estimate it using :code:`pymbar.timeseries.detect_equilibration`.

    Returns
    -------
    preprocessed_data_all : pd.Dataframe
        A list of preprocessed :math:`u_{nk}`/:math:`dH/dλ` data for all replicas that can serve as the
        input to free energy estimators.
    t_list : list
        A list of indices indicating the start of equilibrated data for different replicas. This list will
        be empty if the parameter :code:`t` is specified.
    g_list : list
        A list of statistical inefficiencies of the equilibrated data for different replicas. This list will
        be empty if the parameter :code:`g` is specified.
    """
    if data_type == 'u_nk':
        extract_fn, convert_fn = extract_u_nk, subsampling.u_nk2series
    elif data_type == 'dhdl':
        extract_fn, convert_fn = extract_dHdl, subsampling.dhdl2series
    else:
        raise ValueError("Invalid data_type. Expected 'u_nk' or 'dhdl'.")

    user_specified = None
    if t is None or g is None:
        user_specified = False

    n_sim = len(files_list)
    preprocessed_data_all, t_list, g_list = [], [], []
    for i in range(n_sim):
        print(f'Reading dhdl files of alchemical range {i} ...')
        print(f'Collecting {data_type} data from all iterations ...')
        data = alchemlyb.concat([extract_fn(xvg, T=temp) for xvg in files_list[i]])
        data_series = convert_fn(data)
        data, data_series = subsampling._prepare_input(data, data_series, drop_duplicates=True, sort=True)
        data = subsampling.slicing(data, step=spacing)
        data_series = subsampling.slicing(data_series, step=spacing)

        if user_specified is False:
            print('Estimating the start index of the equilibrated data and the statistical inefficiency ...')
            t, g, Neff_max = detect_equilibration(data_series.values)
            t_list.append(t)
            g_list.append(g)
        else:
            # we only need to estimate Neff_max here.
            Neff_max = int((len(data_series.values) - t) / g)

        print(f'Subsampling and decorrelating the concatenated {data_type} data ...')
        print(f'  Adopted spacing: {spacing: .0f}')
        print(f' {t / len(data_series) * 100: .1f}% of the {data_type} data was in the equilibrium region and therfore discarded.')  # noqa: E501
        print(f'  Statistical inefficiency of {data_type}: {g: .1f}')
        print(f'  Number of effective samples: {Neff_max: .0f}\n')

        data_series_equil, data_equil = data_series[t:], data[t:]
        indices = subsample_correlated_data(data_series_equil, g=g)
        preprocessed_data = data_equil.iloc[indices]

        preprocessed_data_all.append(preprocessed_data)

    return preprocessed_data_all, t_list, g_list


def _apply_estimators(data, df_method="MBAR"):
    """
    An internal function that generates a list of estimators fitting the input data.

    Parameters
    ----------
    data : pd.Dataframe
        A list of dHdl or u_nk dataframes obtained from all replicas of the REXEE simulation of interest.
        Preferrably, the dHdl or u_nk data should be preprocessed by the function proprocess_data.
    df_method : str
        The selected free energy estimator. Options include "MBAR", "BAR" and "TI".

    Returns
    -------
    estimators : list
        A list of estimators fitting the input data for all replicas. With this, the user
        can access all the free energies and their associated uncertainties for all states and replicas.
    """
    n_sim = len(data)
    estimators = []  # A list of objects of the corresponding class in alchemlyb.estimators
    for i in range(n_sim):
        if df_method == "TI":
            estimators.append(TI().fit(data[i]))
        elif df_method == "BAR":
            estimators.append(BAR().fit(data[i]))
        elif df_method == "MBAR":
            estimators.append(MBAR().fit(data[i]))
        else:
            raise ParameterError('Specified estimator not available.')

    return estimators


def _calculate_df_adjacent(estimators):
    """
    An Internal function that calculates at list of free energy between adjacent
    states for all replicas.

    Parameters
    ----------
    estimators : list
        A list of estimators fitting the input data for all replicas. With this, the user
        can access all the free energies and their associated uncertainties for all states and replicas.

    Returns
    -------
    df_adjacent : list
        A list of lists free energy differences between adjacent states for all replicas.
    df_err_adjacent : list
        A list of lists of uncertainties corresponding to the values of :code:`df_adjacent`.
    """
    n_sim = len(estimators)
    df_adjacent = [list(np.array(estimators[i].delta_f_)[:-1, 1:].diagonal()) for i in range(n_sim)]
    df_err_adjacent = [list(np.array(estimators[i].d_delta_f_)[:-1, 1:].diagonal()) for i in range(n_sim)]

    return df_adjacent, df_err_adjacent


def _combine_df_adjacent(df_adjacent, df_err_adjacent, state_ranges, err_type):
    """
    An internal function that combines the free energy differences between adjacent states
    in different state ranges using either simple means or inverse-variance weighted means.
    Specifically, if :code:`df_err_adjacent` is :code:`None`, simple means will be used.
    Otherwise, inverse-variance weighted means will be used.

    Parameters
    ----------
    df_adjacent : list
        A list of lists free energy differences between adjacent states for all replicas.
    df_err_adjacent : list
        A list of lists of uncertainties corresponding to the values of :code:`df_adjacent`.
    state_ranges : list
        A list of lists of intergers that represents the alchemical states that can be sampled by different replicas.
    err_type : str
        How the error of the combined free energy differences should be calculated. Available options include
        "propagate" and "std". Note that the option "propagate" is only available when :code:`df_err_adjacent`
        is not :code:`None`.

    Returns
    -------
    df : list
        A list of free energy differences between states i and i + 1 for the entire state range.
    df_err : list
        A list of uncertainties of the free energy differences for the entire state range.
    overlap_bool : list
        overlap_bool[i] = True means that the i-th free energy difference (i.e. df[i]) was available
        in multiple replicas.
    """
    n_tot = state_ranges[-1][-1] + 1
    df, df_err, overlap_bool = [], [], []
    for i in range(n_tot - 1):
        # df_list is a list of free energy difference between sates i and i+1 in different replicas
        # df_err_list contains the uncertainties corresponding to the values of df_list
        df_list, df_err_list = [], []
        for j in range(len(state_ranges)):   # len(state_ranges) = n_sim
            if i in state_ranges[j] and i + 1 in state_ranges[j]:
                idx = state_ranges[j].index(i)
                df_list.append(df_adjacent[j][idx])
                if df_err_adjacent is not None:
                    df_err_list.append(df_err_adjacent[j][idx])
        overlap_bool.append(len(df_list) > 1)

        if df_err_adjacent is None:
            # simple means and std will be used
            mean, error = np.mean(df_list), np.std(df_list, ddof=1)
        else:
            # inverse-variance weighted means and propagated error will be used
            mean, error = utils.weighted_mean(df_list, df_err_list)

            if err_type == 'std':
                # overwrite the error calculated above
                error = np.std(df_list, ddof=1)

        df.append(mean)
        df_err.append(error)

    return df, df_err, overlap_bool


def calculate_free_energy(data, state_ranges, df_method="MBAR", err_method='propagate', n_bootstrap=None, seed=None):
    """
    Caculates the averaged free energy profile with the chosen method given dHdl or u_nk data obtained from
    all replicas of the REXEE simulation of interest. Available methods include TI, BAR, and MBAR. TI
    requires dHdl data while the other two require u_nk data.

    Parameters
    ----------
    data : pd.Dataframe
        A list of dHdl or u_nk dataframes obtained from all replicas of the REXEE simulation of interest.
        Preferrably, the dHdl or u_nk data should be preprocessed by the function proprocess_data.
    state_ranges : list
        A list of lists of intergers that represents the alchemical states that can be sampled by different replicas.
    df_method : str
        The method used to calculate the free energy profile. Available choices include "TI", "BAR", and "MBAR".
    err_method : str
        The method used to estimate the uncertainty of the free energy combined across multiple replicas.
        Available options include "propagate" and "bootstrap". The bootstrapping method is more accurate
        but much more computationally expensive than simple error propagation.
    n_bootstrap : int
        The number of bootstrap iterations. This parameter is used only when the boostrapping method is chosen to
        estimate the uncertainties of the free energies.
    seed : int
        The random seed for bootstrapping.

    Returns
    -------
    f : list
        The full-range free energy profile.
    f_err : list
        The uncertainty corresponding to the values in :code:`f`.
    estimators : list
        A list of estimators fitting the input data for all replicas. With this, the user
        can access all the free energies and their associated uncertainties for all states and replicas.
    """
    n_sim = len(data)
    n_tot = state_ranges[-1][-1] + 1
    estimators = _apply_estimators(data, df_method)
    df_adjacent, df_err_adjacent = _calculate_df_adjacent(estimators)
    df, df_err, overlap_bool = _combine_df_adjacent(df_adjacent, df_err_adjacent, state_ranges, err_type='propagate')

    if err_method == 'bootstrap':
        if seed is not None:
            print(f'Setting the random seed for boostrapping: {seed}')

        # Recalculate err with bootstrapping. (df is still the same and has been calculated above.)
        df_bootstrap = []
        sampled_data_all = [data[i].sample(n=len(data[i]) * n_bootstrap, replace=True, random_state=seed) for i in range(n_sim)]  # noqa: E501
        for b in range(n_bootstrap):
            sampled_data = [sampled_data_all[i].iloc[b * len(data[i]):(b + 1) * len(data[i])] for i in range(n_sim)]
            bootstrap_estimators = _apply_estimators(sampled_data, df_method)
            df_adjacent, df_err_adjacent = _calculate_df_adjacent(bootstrap_estimators)
            df_sampled, _, overlap_bool = _combine_df_adjacent(df_adjacent, df_err_adjacent, state_ranges, err_type='propagate')  # doesn't matter what value err_type here is # noqa: E501
            df_bootstrap.append(df_sampled)
        error_bootstrap = np.std(df_bootstrap, axis=0, ddof=1)

        # Replace the value in df_err with value in error_bootstrap if df_err corresponds to
        # the df between overlapping states
        for i in range(n_tot - 1):
            if overlap_bool[i] is True:
                print(f'Replaced the propagated error with the bootstrapped error for states {i} and {i + 1}: {df_err[i]:.5f} -> {error_bootstrap[i]:.5f}.')  # noqa: E501
                df_err[i] = error_bootstrap[i]

    df.insert(0, 0)
    df_err.insert(0, 0)
    f = [sum(df[:(i + 1)]) for i in range(len(df))]
    f_err = [np.sqrt(sum([x**2 for x in df_err[:(i+1)]])) for i in range(len(df_err))]

    return f, f_err, estimators


def calculate_df_rmse(estimators, df_ref, state_ranges):
    """
    Calculates the RMSE values of the free energy profiles of different alchemical ranges given the reference free
    energy profile for the whole range of states.

    Parameters
    ----------
    estimators : list
        A list of estimators fitting the input data for all replicas. With this, the user
        can access all the free energies and their associated uncertainties for all states and replicas.
    df_ref : list
        A list of values corresponding to the free energies of the whole range of states. The length
        of the list should be equal to the number of states in total.
    state_ranges : list
        A list of lists of intergers that represents the alchemical states that can be sampled by different replicas.

    Returns
    -------
    rmse_list : list
        A list of RMSE values of the free energy profiles of different alchemical ranges.
    """
    n_sim = len(estimators)
    df_ref = np.array(df_ref)
    rmse_list = []
    for i in range(n_sim):
        df = np.array(estimators[i].delta_f_.iloc[0])  # the first state always has 0 free energy here
        ref = df_ref[state_ranges[i]]
        ref -= ref[0]   # shift the free energy of the first state in the range to 0
        rmse_list.append(np.sqrt(np.sum((df - ref) ** 2) / len(df)))

    return rmse_list


def plot_free_energy(f, f_err, fig_name):
    """
    Plot the free energy profile with error bars.

    Parameters
    ----------
    f : list
        The full-range free energy profile.
    f_err : list
        The uncertainty corresponding to the values in :code:`f`.
    fig_name : str
        The file name of the png file to be saved (with the extension).
    """
    plt.figure()
    plt.plot(range(len(f)), f, 'o-', c='#1f77b4')
    plt.errorbar(range(len(f)), f, yerr=f_err, fmt='o', capsize=2, c='#1f77b4')
    plt.xlabel('State')
    plt.ylabel('Free energy (kT)')
    plt.grid()
    plt.savefig(f'{fig_name}', dpi=600)


def average_weights(g_vecs, frac):
    """
    Average the differences between the weights of the coupled and uncoupled states.
    This can be an estimate of the free energy difference between two end states.

    Parameters
    ----------
    g_vecs : np.array
        An array of alchemical weights of the whole range of states as a function of
        simulation time, which is typically generated by :obj:`.combine_weights`.
    frac : float
        The fraction of g_vecs to average over. frac=0.2 means average the last 20% of
        the weight vectors.

    Returns
    -------
    dg_avg : float
        The averaged difference in the weights between the coupled and uncoupled states.
    dg_avg_err : float
        The error of :code:`dg_avg`.
    """
    N = len(g_vecs)
    dg = []
    for i in range(N):
        dg.append(g_vecs[i][-1] - g_vecs[i][0])
    n = int(np.floor(N * frac))
    dg_avg = np.mean(dg[-n:])
    dg_avg_err = np.std(dg_avg[-n:], ddof=1)

    return dg_avg, dg_avg_err



####################################################################
#                                                                  #
#    ensemble_md,                                                  #
#    a python package for running GROMACS simulation ensembles     #
#                                                                  #
#    Written by Wei-Tse Hsu <wehs7661@colorado.edu>                #
#    Copyright (c) 2022 University of Colorado Boulder             #
#                                                                  #
####################################################################
import os
import sys
import time
import copy
import shutil
import traceback
import argparse
import numpy as np
from mpi4py import MPI
from datetime import datetime

from ensemble_md.utils import utils
from ensemble_md.replica_exchange_EE import ReplicaExchangeEE


def initialize(args):
    parser = argparse.ArgumentParser(
        description='This code runs a REXEE simulation given necessary inputs.')
    parser.add_argument('-y',
                        '--yaml',
                        type=str,
                        default='params.yaml',
                        help='The input YAML file that contains REXEE parameters. (Default: params.yaml)')
    parser.add_argument('-c',
                        '--ckpt',
                        type=str,
                        default='rep_trajs.npy',
                        help='The NPY file containing the replica-space trajectories. This file is a \
                            necessary checkpoint file for extending the simulaiton. (Default: rep_trajs.npy)')
    parser.add_argument('-g',
                        '--g_vecs',
                        type=str,
                        default='g_vecs.npy',
                        help='The NPY file containing the timeseries of the whole-range alchemical weights. \
                            This file is a necessary input if ones wants to update the file when extending \
                            the simulation. (Default: g_vecs.npy)')
    parser.add_argument('-o',
                        '--output',
                        type=str,
                        default='run_REXEE_log.txt',
                        help='The output file for logging how replicas interact with each other. \
                            (Default: run_REXEE_log.txt)')
    parser.add_argument('-m',
                        '--maxwarn',
                        type=int,
                        default=0,
                        help='The maximum number of warnings in parameter specification to be ignored.')
    args_parse = parser.parse_args(args)

    return args_parse


def main():
    t1 = time.time()
    args = initialize(sys.argv[1:])
    sys.stdout = utils.Logger(logfile=args.output)
    sys.stderr = utils.Logger(logfile=args.output)

    # Step 1: Set up MPI rank and instantiate ReplicaExchangeEE to set up REXEE parameters
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()  # Note that this is a GLOBAL variable

    if rank == 0:
        print(f'Current time: {datetime.now().strftime("%d/%m/%Y %H:%M:%S")}')
        print(f'Command line: {" ".join(sys.argv)}\n')

    REXEE = ReplicaExchangeEE(args.yaml)

    if rank == 0:
        # Print out simulation parameters
        REXEE.print_params()

        # Print out warnings and fail if needed
        for i in REXEE.warnings:
            print(f'\n{i}\n')

        if len(REXEE.warnings) > args.maxwarn:
            print(f"The execution failed due to warning(s) about parameter spcificaiton. Check the warnings, or consider setting maxwarn in the input YAML file if you find them harmless.")  # noqa: E501, F541
            comm.Abort(101)

    # Step 2: If there is no checkpoint file found/provided, perform the 1st iteration (index 0)
    if os.path.isfile(args.ckpt) is False:
        start_idx = 1

        # 2-1. Set up input files for all simulations
        if rank == 0:
            for i in range(REXEE.n_sim):
                os.mkdir(f'sim_{i}')
                os.mkdir(f'sim_{i}/iteration_0')
                MDP = REXEE.initialize_MDP(i)
                MDP.write(f"sim_{i}/iteration_0/expanded.mdp", skipempty=True)

        # 2-2. Run the first set of simulations
        REXEE.run_REXEE(0)

    else:
        if rank == 0:
            # If there is a checkpoint file, we see the execution as an extension of an REXEE simulation
            ckpt_data = np.load(args.ckpt)
            start_idx = len(ckpt_data[0])  # The length should be the same for the same axis
            print(f'\nGetting prepared to extend the REXEE simulation from iteration {start_idx} ...')

            if start_idx == REXEE.n_iter:
                print('Extension aborted: The expected number of iterations have been completed!')
                MPI.COMM_WORLD.Abort(1)
            else:
                print('Deleting data generated after the checkpoint ...')
                for i in range(REXEE.n_sim):
                    n_finished = len(next(os.walk(f'sim_{i}'))[1])  # number of finished iterations
                    for j in range(start_idx, n_finished):
                        print(f'  Deleting the folder sim_{i}/iteration_{j}')
                        shutil.rmtree(f'sim_{i}/iteration_{j}')

            # Read g_vecs.npy and rep_trajs.npy so that new data can be appended, if any.
            # Note that these two arrays are created in rank 0 and should always be operated in rank 0,
            # or broadcasting is required.
            REXEE.rep_trajs = [list(i) for i in ckpt_data]
            if os.path.isfile(args.g_vecs) is True:
                REXEE.g_vecs = [list(i) for i in np.load(args.g_vecs)]
        else:
            start_idx = None

        start_idx = comm.bcast(start_idx, root=0)  # so that all the ranks are aware of start_idx

    # 2-3. Get the reference distance for the distance restraint specified in the pull code, if any.
    REXEE.get_ref_dist()

    for i in range(start_idx, REXEE.n_iter):
        # For a large code block like below executed on rank 0, we try to catch any exception and abort the simulation.
        # So if there is bug, the execution will be terminated and no computation time will be wasted.
        try:
            if rank == 0:
                # Step 3: Swap the coordinates
                # 3-1. For all the replica simulations,
                #   (1) Find the last sampled state and the corresponding lambda values from the DHDL files.
                #   (2) Find the final Wang-Landau incrementors and weights from the LOG files.
                dhdl_files = [f'sim_{j}/iteration_{i - 1}/dhdl.xvg' for j in range(REXEE.n_sim)]
                log_files = [f'sim_{j}/iteration_{i - 1}/md.log' for j in range(REXEE.n_sim)]
                states_ = REXEE.extract_final_dhdl_info(dhdl_files)
                wl_delta, weights_, counts_ = REXEE.extract_final_log_info(log_files)
                print()

                # 3-2. Identify swappable pairs, propose swap(s), calculate P_acc, and accept/reject swap(s)
                # Note after `get_swapping_pattern`, `states_` and `weights_` won't be necessarily
                # since they are updated by `get_swapping_pattern`. (Even if the function does not explicitly
                # returns `states_` and `weights_`, `states_` and `weights_` can still be different after
                # the use of the function.) Therefore, here we create copies for `states_` and `weights_`
                # before the use of `get_swapping_pattern`, so we can use them in `weight_correction`,
                # `combine_weights` and `update_MDP`.
                states = copy.deepcopy(states_)
                weights = copy.deepcopy(weights_)
                counts = copy.deepcopy(counts_)
                swap_pattern, swap_list = REXEE.get_swapping_pattern(dhdl_files, states_)  # swap_list will only be used for modify_coords  # noqa: E501

                # 3-3. Perform weight correction/weight combination
                if wl_delta != [None for i in range(REXEE.n_sim)]:  # weight-updating
                    print(f'\nCurrent Wang-Landau incrementors: {wl_delta}\n')

                # (1) First we prepare the time-averaged weights to be combined, if needed.
                # Note that although averaged weights are sometimes used for weight correction/weight combination,
                # the final weights are always used for calculating the acceptance ratio.
                if REXEE.N_cutoff != -1 or REXEE.w_combine is True:
                    # Only when weight correction/weight combination is needed.
                    weights_avg, weights_err = REXEE.get_averaged_weights(log_files)

                    # Calculate the RMSE between the averaged weights and the final weights by the way.
                    rmse_list = [utils.calc_rmse(weights_avg[i], weights[i]) for i in range(REXEE.n_sim)]
                    rmse_str = ', '.join([f'{i:.2f}' for i in rmse_list])
                    print(f'RMSE between the final weights and time-averaged weights for each replica: {rmse_str} kT')

                # (2) Now we perform weight correction/weight combination.
                # The product of this step should always be named as "weights" to be used in update_MDP
                if REXEE.N_cutoff != -1 and REXEE.w_combine is True:
                    # Perform both weight correction and weight combination
                    if REXEE.verbose is True:
                        print('Performing weight correction ...')
                    else:
                        print('Performing weight correction ...', end='')
                    weights_preprocessed = REXEE.weight_correction(weights_avg, counts)

                    if REXEE.verbose is True:
                        print('Performing weight combination ...')
                    else:
                        print('Performing weight combination ...', end='')
                    if REXEE.w_mean_type == 'simple':
                        weights, g_vec = REXEE.combine_weights(weights_preprocessed)  # simple means
                    else:
                        # Note that here weights_err are acutally not the uncertainties for weights_prepocessed
                        # but weights_avg ... We might need to disable this feature in the future.
                        weights, g_vec = REXEE.combine_weights(weights_preprocessed, weights_err)  # inverse-variance weighting  # noqa: E501
                    REXEE.g_vecs.append(g_vec)

                    # Check if histogram correction is needed after weight combination
                    if REXEE.hist_corr is True:
                        print('Performing histogram correction ...')
                        counts = REXEE.histogram_correction(counts_)
                    else:
                        print('Note: No histogram correction will be performed.')

                elif REXEE.N_cutoff == -1 and REXEE.w_combine is True:
                    # Only perform weight combination
                    print('Note: No weight correction will be performed.')
                    if REXEE.verbose is True:
                        print('Performing weight combination ...')
                    else:
                        print('Performing weight combination ...', end='')
                    if REXEE.w_mean_type == 'simple':
                        weights, g_vec = REXEE.combine_weights(weights_avg)  # simple means
                    else:
                        weights, g_vec = REXEE.combine_weights(weights_avg, weights_err)  # inverse-variance weighting
                    REXEE.g_vecs.append(g_vec)

                    # Check if histogram correction is needed after weight combination
                    if REXEE.hist_corr is True:
                        print('Performing histogram correction ...')
                        counts = REXEE.histogram_correction(counts_)
                    else:
                        print('Note: No histogram correction will be performed.')

                elif REXEE.N_cutoff != -1 and REXEE.w_combine is False:
                    # Only perform weight correction
                    print('Note: No weight combination will be performed.')
                    if REXEE.verbose is True:
                        print('Performing weight correction ...')
                    else:
                        print('Performing weight correction ...', end='')
                    weights = REXEE.weights_correction(weights_avg, counts)
                    _ = REXEE.combine_weights(weights, print_values=False)[1]  # just to print the combined weights  # noqa: E501
                else:
                    print('Note: No weight correction will be performed.')
                    print('Note: No weight combination will be performed.')
                    # Note that in this case, the final weights will be used in the next iteration.
                    _ = REXEE.combine_weights(weights, print_values=False)[1]  # just to print the combiend weights  # noqa: E501

                # 3-4. Modify the MDP files and swap out the GRO files (if needed)
                # Here we keep the lambda range set in mdp the same across different iterations in the same folder but swap out the gro file  # noqa: E501
                # Note we use states (copy of states_) instead of states_ in update_MDP.
                for j in list(range(REXEE.n_sim)):
                    os.mkdir(f'sim_{j}/iteration_{i}')
                    MDP = REXEE.update_MDP(f"sim_{j}/iteration_{i - 1}/expanded.mdp", j, i, states, wl_delta, weights, counts)   # modify with a new template  # noqa: E501
                    MDP.write(f"sim_{j}/iteration_{i}/expanded.mdp", skipempty=True)
                    # In run_REXEE(i, swap_pattern), where the tpr files will be generated, we use the top file at the
                    # level of the simulation (the file that will be shared by all simulations). For the gro file, we
                    # pass swap_pattern to the function to figure it out internally.
            else:
                swap_pattern, swap_list = None, None

        except Exception:
            print('\n--------------------------------------------------------------------------\n')
            print(f'An error occurred on rank 0:\n{traceback.format_exc()}')
            MPI.COMM_WORLD.Abort(1)

        if -1 not in REXEE.equil and 0 not in REXEE.equil:
            # This is the case where the weights are equilibrated in a weight-updating simulation.
            # As a remidner, REXEE.equil should be a list of 0 after extract_final_log_info in a
            # fixed-weight simulation, and a list of -1 for a weight-updating simulation with unequilibrated weights.
            print('\nSimulation terminated: The weights have been equilibrated for all replicas.')  # this will only be printed in rank 0  # noqa: E501

            # Note that REXEE.equil is avaiable for all ranks but only updated in rank 0. So the if condition here
            # can only be satisfied in rank 0. We broadcast exit_loop to all ranks so that all ranks can exit the
            # simulation at the same time, if the weights get equilibrated.
            exit_loop = True
            comm.bcast(exit_loop, root=0)
            break
        else:
            # Receive exit_loop from rank 0. If exit_loop is True, all ranks will exit the simulation.
            exit_loop = comm.bcast(False, root=0)
            if exit_loop:
                break

        # Step 4: Perform another iteration
        # 4-1. Modify the coordinates of the output gro files if needed.
        swap_pattern = comm.bcast(swap_pattern, root=0)
        swap_list = comm.bcast(swap_list, root=0)

        if len(swap_list) == 0:
            pass
        else:
            if REXEE.modify_coords_fn is not None:
                try:
                    if rank == 0:
                        for j in range(len(swap_list)):
                            print('\nModifying the coordinates of the following output GRO files ...')
                            # gro_1 and gro_2 are the simlation outputs (that we want to back up) and the inputs to modify_coords  # noqa: E501
                            gro_1 = f'sim_{swap_list[j][0]}/iteration_{i-1}/confout.gro'
                            gro_2 = f'sim_{swap_list[j][1]}/iteration_{i-1}/confout.gro'
                            print(f'  - {gro_1}\n  - {gro_2}')

                            # Now we rename gro_1 and gro_2 to back them up
                            gro_1_backup = gro_1.split('.gro')[0] + '_backup.gro'
                            gro_2_backup = gro_2.split('.gro')[0] + '_backup.gro'
                            os.rename(gro_1, gro_1_backup)
                            os.rename(gro_2, gro_2_backup)

                            # Here we input gro_1_backup and gro_2_backup and modify_coords_fn will save the modified gro files as gro_1 and gro_2  # noqa: E501
                            REXEE.modify_coords_fn(gro_1_backup, gro_2_backup)  # the order should not matter
                except Exception:
                    print('\n--------------------------------------------------------------------------\n')
                    print(f'\nAn error occurred on rank 0:\n{traceback.format_exc()}')
                    MPI.COMM_WORLD.Abort(1)

        # 4-2. Run another set of simulations
        REXEE.run_REXEE(i, swap_pattern)

        # 4-3. Save data
        if rank == 0:
            if (i + 1) % REXEE.n_ckpt == 0:
                if len(REXEE.g_vecs) != 0:
                    # Save g_vec as a function of time if weight combination was used.
                    np.save('g_vecs.npy', REXEE.g_vecs)

                print('\n----- Saving .npy files to checkpoint the simulation ---')
                np.save('rep_trajs.npy', REXEE.rep_trajs)

    # Save the npy files at the end of the simulation anyway.
    if rank == 0:
        if len(REXEE.g_vecs) != 0:  # The length will be 0 only if there is no weight combination.
            np.save('g_vecs.npy', REXEE.g_vecs)
        np.save('rep_trajs.npy', REXEE.rep_trajs)

    # Step 5: Write a summary for the simulation ensemble
    if rank == 0:
        print('\nSummary of the simulation ensemble')
        print('==================================')
        print('Simulation status:')
        for i in range(REXEE.n_sim):
            if REXEE.fixed_weights is True:
                print(f'- Rep {i}: The weights were fixed throughout the simulation.')
            elif REXEE.equil[i] == -1:
                print(f'  - Rep {i}: The weights have not been equilibrated.')
            else:
                idx = int(np.floor(REXEE.equil[i] / (REXEE.dt * REXEE.nst_sim)))
                if REXEE.equil[i] > 1000:
                    units = 'ns'
                    REXEE.equil[i] /= 1000
                else:
                    units = 'ps'
                print(f'  - Rep {i}: The weights have been equilibrated at {REXEE.equil[i]:.2f} {units} (iteration {idx}).')  # noqa: E501

        print(f'\n{REXEE.n_empty_swappable} out of {REXEE.n_iter}, or {REXEE.n_empty_swappable / REXEE.n_iter * 100:.1f}% iterations had an empty list of swappable pairs.')  # noqa: E501
        if REXEE.n_swap_attempts != 0:
            print(f'{REXEE.n_rejected} out of {REXEE.n_swap_attempts}, or {REXEE.n_rejected / REXEE.n_swap_attempts * 100:.1f}% of attempted exchanges were rejected.')  # noqa: E501

        print(f'\nTime elapsed: {utils.format_time(time.time() - t1)}')

    MPI.Finalize()



####################################################################
#                                                                  #
#    ensemble_md,                                                  #
#    a python package for running GROMACS simulation ensembles     #
#                                                                  #
#    Written by Wei-Tse Hsu <wehs7661@colorado.edu>                #
#    Copyright (c) 2022 University of Colorado Boulder             #
#                                                                  #
####################################################################
"""
Unit tests for the module utils.py.
"""
import sys
import tempfile
import numpy as np
from ensemble_md.utils import utils


def test_logger():
    # Create a temporary file for the log
    with tempfile.TemporaryFile(mode="w+t") as log_file:
        # Get the file path for the temporary file
        log_path = log_file.name

        # Create a logger that redirects output to the temporary file
        logger = utils.Logger(log_path)

        # Redirect stdout to the logger
        sys.stdout = logger

        # Write some messages to stdout
        print("Hello, world!")
        print("Testing logger...")

        # Flush the logger to ensure that all messages are written to the log
        logger.flush()

        # Reset stdout to the original stream
        sys.stdout = sys.__stdout__


def test_format_time():
    assert utils.format_time(0) == "0.0 second(s)"
    assert utils.format_time(1) == "1.0 second(s)"
    assert utils.format_time(59) == "59.0 second(s)"
    assert utils.format_time(60) == "1 minute(s) 0 second(s)"
    assert utils.format_time(61) == "1 minute(s) 1 second(s)"
    assert utils.format_time(3599) == "59 minute(s) 59 second(s)"
    assert utils.format_time(3600) == "1 hour(s) 0 minute(s) 0 second(s)"
    assert utils.format_time(3661) == "1 hour(s) 1 minute(s) 1 second(s)"
    assert utils.format_time(86399) == "23 hour(s) 59 minute(s) 59 second(s)"
    assert utils.format_time(86400) == "1 day, 0 hour(s) 0 minute(s) 0 second(s)"
    assert utils.format_time(90061) == "1 day, 1 hour(s) 1 minute(s) 1 second(s)"


def test_autoconvert():
    # Test non-string input
    assert utils._autoconvert(42) == 42

    # Test string input that can be converted to int
    assert utils._autoconvert("42") == 42

    # Test string input that can be converted to float
    assert utils._autoconvert("3.14159") == 3.14159

    # Test string input that can be converted to a numpy array of ints
    assert utils._autoconvert("1 2 3") == [1, 2, 3]

    # Test string input that can be converted to a numpy array of floats
    assert utils._autoconvert("1.0 2.0 3.0") == [1.0, 2.0, 3.0]


def test_get_subplot_dimension():
    assert utils._get_subplot_dimension(1) == (1, 1)
    assert utils._get_subplot_dimension(2) == (1, 2)
    assert utils._get_subplot_dimension(3) == (2, 2)
    assert utils._get_subplot_dimension(4) == (2, 2)
    assert utils._get_subplot_dimension(5) == (2, 3)
    assert utils._get_subplot_dimension(6) == (2, 3)
    assert utils._get_subplot_dimension(7) == (3, 3)
    assert utils._get_subplot_dimension(8) == (3, 3)
    assert utils._get_subplot_dimension(9) == (3, 3)


def test_weighted_mean():
    # 1. Same error for each sample --> weighted mean reduces to simple mean
    vals = [1, 2, 3, 4]
    errs = [0.1, 0.1, 0.1, 0.1]
    mean, err = utils.weighted_mean(vals, errs)
    assert mean == 2.5
    assert err == np.sqrt(4 * 0.1 ** 2) / 4  # 0.05

    # 2. Different errors
    vals = [1, 2, 3, 4]
    errs = [5, 0.1, 0.1, 0.1]
    mean, err = utils.weighted_mean(vals, errs)
    assert np.isclose(mean, 2.9997333688841485)
    assert np.isclose(err, 0.0577311783020254)



####################################################################
#                                                                  #
#    ensemble_md,                                                  #
#    a python package for running GROMACS simulation ensembles     #
#                                                                  #
#    Written by Wei-Tse Hsu <wehs7661@colorado.edu>                #
#    Copyright (c) 2022 University of Colorado Boulder             #
#                                                                  #
####################################################################
"""
Unit tests for the module replica_exchange_EE.py.
"""
import os
import sys
import yaml
import copy
import random
import pytest
import numpy as np
import ensemble_md
from ensemble_md.utils import gmx_parser
from ensemble_md.replica_exchange_EE import ReplicaExchangeEE
from ensemble_md.utils.exceptions import ParameterError

current_path = os.path.dirname(os.path.abspath(__file__))
input_path = os.path.join(current_path, "data")


@pytest.fixture
def params_dict():
    """
    Generates a dictionary containing the required REXEE parameters.
    """
    REXEE_dict = {
        'gmx_executable': 'gmx',
        'gro': 'ensemble_md/tests/data/sys.gro',
        'top': 'ensemble_md/tests/data/sys.top',
        'mdp': 'ensemble_md/tests/data/expanded.mdp',
        'n_sim': 4,
        'n_iter': 10,
        's': 1,
    }
    yield REXEE_dict

    # Remove the file after the unit test is done.
    if os.path.isfile('params.yaml') is True:
        os.remove('params.yaml')


def get_REXEE_instance(input_dict, yml_file='params.yaml'):
    """
    Saves a dictionary as a yaml file and use it to instantiate the ReplicaExchangeEE class.
    """
    with open(yml_file, 'w') as f:
        yaml.dump(input_dict, f)
    REXEE = ReplicaExchangeEE(yml_file)
    return REXEE


def check_param_error(REXEE_dict, param, match, wrong_val='cool', right_val=None):
    """
    Test if ParameterError is raised if a parameter is not well-defined.
    """
    REXEE_dict[param] = wrong_val
    with pytest.raises(ParameterError, match=match):
        get_REXEE_instance(REXEE_dict)
    REXEE_dict[param] = right_val  # so that REXEE_dict can be read without failing in later assertions

    return REXEE_dict


class Test_ReplicaExchangeEE:
    def test_init(self, params_dict):
        REXEE = get_REXEE_instance(params_dict)
        assert REXEE.yaml == 'params.yaml'

    def test_set_params_error(self, params_dict):
        # 1. Required parameters
        del params_dict['n_sim']
        with pytest.raises(ParameterError, match="Required parameter 'n_sim' not specified in params.yaml."):
            get_REXEE_instance(params_dict)
        params_dict['n_sim'] = 4  # so params_dict can be read without failing in the assertions below

        # 2. Available options
        check_param_error(params_dict, 'proposal', "The specified proposal scheme is not available. Available options include 'single', 'neighboring', and 'exhaustive'.", 'cool', 'exhaustive')  # noqa: E501
        check_param_error(params_dict, 'df_method', "The specified free energy estimator is not available. Available options include 'TI', 'BAR', and 'MBAR'.")  # noqa: E501
        check_param_error(params_dict, 'err_method', "The specified method for error estimation is not available. Available options include 'propagate', and 'bootstrap'.")  # noqa: E501

        # 3. Integer parameters
        check_param_error(params_dict, 'nst_sim', "The parameter 'nst_sim' should be an integer.")
        check_param_error(params_dict, 'seed', "The parameter 'seed' should be an integer.")
        check_param_error(params_dict, 'n_sim', "The parameter 'n_sim' should be an integer.", 4.1, 4)

        # 4. Positive parameters
        check_param_error(params_dict, 'nst_sim', "The parameter 'nst_sim' should be positive.", -5, 500)
        check_param_error(params_dict, 'n_iter', "The parameter 'n_iter' should be positive.", 0, 10)

        # 5. Non-negative parameters
        check_param_error(params_dict, 'N_cutoff', "The parameter 'N_cutoff' should be non-negative unless no weight correction is needed, i.e. N_cutoff = -1.", -5)  # noqa: E501

        # 6. String parameters
        check_param_error(params_dict, 'mdp', "The parameter 'mdp' should be a string.", 3, 'ensemble_md/tests/data/expanded.mdp')  # noqa: E501

        # 7. Boolean parameters
        check_param_error(params_dict, 'msm', "The parameter 'msm' should be a boolean variable.", 3, False)

        # 8. nstlog > nst_sim
        mdp = gmx_parser.MDP(os.path.join(input_path, "expanded.mdp"))  # A perfect mdp file
        mdp['nstlog'] = 200
        mdp.write(os.path.join(input_path, "expanded_test.mdp"))
        params_dict['mdp'] = 'ensemble_md/tests/data/expanded_test.mdp'
        params_dict['nst_sim'] = 100
        with pytest.raises(ParameterError, match='The parameter "nstlog" must be a factor of the parameter "nst_sim" specified in the YAML file.'):  # noqa: E501
            get_REXEE_instance(params_dict)
        params_dict['nst_sim'] = 500
        os.remove(os.path.join(input_path, "expanded_test.mdp"))

        # 9. n_sub < 1
        params_dict['s'] = 5
        params_dict['mdp'] = 'ensemble_md/tests/data/expanded.mdp'
        # Note that the parentheses are special characters that need to be escaped in regular expressions
        with pytest.raises(ParameterError, match=r'There must be at least two states for each replica \(current value: -6\). The current specified configuration \(n_tot=9, n_sim=4, s=5\) does not work for REXEE.'):  # noqa: E501
            get_REXEE_instance(params_dict)
        params_dict['s'] = 1

    def test_set_params_warnings(self, params_dict):
        # 1. Non-recognizable parameter in the YAML file
        params_dict['cool'] = 10
        REXEE = get_REXEE_instance(params_dict)
        warning = 'Warning: Parameter "cool" specified in the input YAML file is not recognizable.'
        assert warning in REXEE.warnings

        # 2. Warnings related to the mdp file
        mdp = gmx_parser.MDP(os.path.join(input_path, "expanded.mdp"))  # A perfect mdp file
        mdp['lmc_seed'] = 1000
        mdp['gen_seed'] = 1000
        mdp['wl_scale'] = ''
        mdp.write(os.path.join(input_path, "expanded_test.mdp"))

        params_dict['mdp'] = 'ensemble_md/tests/data/expanded_test.mdp'
        params_dict['N_cutoff'] = 1000
        REXEE = get_REXEE_instance(params_dict)

        warning_1 = 'Warning: The weight correction/weight combination method is specified but will not be used since the weights are fixed.'  # noqa: E501
        warning_2 = 'Warning: We recommend setting lmc_seed as -1 so the random seed is different for each iteration.'
        warning_3 = 'Warning: We recommend setting gen_seed as -1 so the random seed is different for each iteration.'
        assert warning_1 in REXEE.warnings
        assert warning_2 in REXEE.warnings
        assert warning_3 in REXEE.warnings

        os.remove(os.path.join(input_path, "expanded_test.mdp"))

    def test_set_params(self, params_dict):
        # 0. Get an REXEE instance to test
        REXEE = get_REXEE_instance(params_dict)

        # 1. Check the required REXEE parameters
        assert REXEE.gmx_executable == 'gmx'
        assert REXEE.gro == "ensemble_md/tests/data/sys.gro"
        assert REXEE.top == "ensemble_md/tests/data/sys.top"
        assert REXEE.mdp == "ensemble_md/tests/data/expanded.mdp"
        assert REXEE.n_sim == 4
        assert REXEE.n_iter == 10
        assert REXEE.s == 1

        # 2. Check the default values of the parameters not specified in params.yaml
        assert REXEE.proposal == "exhaustive"
        assert REXEE.w_combine is False
        assert REXEE.N_cutoff == 1000
        assert REXEE.verbose is True
        assert REXEE.runtime_args is None
        assert REXEE.n_ckpt == 100
        assert REXEE.msm is False
        assert REXEE.free_energy is False
        assert REXEE.df_spacing == 1
        assert REXEE.df_method == 'MBAR'
        assert REXEE.err_method == 'propagate'
        assert REXEE.n_bootstrap == 50
        assert REXEE.seed is None

        # 3. Check the MDP parameters
        assert REXEE.dt == 0.002
        assert REXEE.temp == 298
        assert REXEE.nst_sim == 500
        assert REXEE.fixed_weights is False

        # 4. Checked the derived parameters
        # Note that lambda_dict will also be tested in test_map_lambda2state.
        k = 1.380649e-23
        NA = 6.0221408e23
        assert REXEE.kT == k * NA * 298 / 1000
        assert REXEE.lambda_types == ['coul_lambdas', 'vdw_lambdas']
        assert REXEE.n_tot == 9
        assert REXEE.n_sub == 6
        assert REXEE.state_ranges == [
            [0, 1, 2, 3, 4, 5],
            [1, 2, 3, 4, 5, 6],
            [2, 3, 4, 5, 6, 7],
            [3, 4, 5, 6, 7, 8], ]
        assert REXEE.equil == [-1, -1, -1, -1]  # will be zeros after the first iteration if the weights are fixed
        assert REXEE.n_rejected == 0
        assert REXEE.n_swap_attempts == 0
        assert REXEE.n_empty_swappable == 0
        assert REXEE.rep_trajs == [[0], [1], [2], [3]]

        params_dict['df_method'] = 'MBAR'
        REXEE = get_REXEE_instance(params_dict)
        assert REXEE.df_data_type == 'u_nk'

        params_dict['df_method'] = 'BAR'
        REXEE = get_REXEE_instance(params_dict)
        assert REXEE.df_data_type == 'dhdl'

        params_dict['grompp_args'] = {'-maxwarn': 1}
        params_dict['runtime_args'] = {'-nt': 16, '-ntomp': 8}
        REXEE = get_REXEE_instance(params_dict)
        assert REXEE.runtime_args == {'-nt': 16, '-ntomp': 8}
        assert REXEE.grompp_args == {'-maxwarn': 1}

        assert REXEE.reformatted_mdp is False

    def test_set_params_edge_cases(self, params_dict):
        # In the previous unit tests, we have tested the case where fixed_weights is False. Here we test the others.
        mdp = gmx_parser.MDP(os.path.join(input_path, "expanded.mdp"))

        # 1. wl_scale is not specified
        del mdp['wl_scale']
        mdp.write(os.path.join(input_path, "expanded_test.mdp"))
        params_dict['mdp'] = 'ensemble_md/tests/data/expanded_test.mdp'
        REXEE = get_REXEE_instance(params_dict)
        assert REXEE.fixed_weights is True

        # 2. wl_scale is not empty
        mdp['wl_scale'] = ''  # It will become an empty array array([]) after read by gmx_parser.abs(x)
        mdp.write(os.path.join(input_path, "expanded_test.mdp"))
        REXEE = get_REXEE_instance(params_dict)
        assert REXEE.fixed_weights is True

    def test_reformat_MDP(self, params_dict):
        # Note that the function reformat_MDP is called in set_params
        mdp = gmx_parser.MDP(os.path.join(input_path, "expanded.mdp"))
        mdp['cool-stuff'] = 10
        mdp.write(os.path.join(input_path, "expanded_test.mdp"))
        params_dict['mdp'] = 'ensemble_md/tests/data/expanded_test.mdp'
        REXEE = get_REXEE_instance(params_dict)  # this calss reformat_MDP internally
        assert REXEE.reformatted_mdp is True
        assert REXEE.template['cool_stuff'] == 10
        assert os.path.isfile(os.path.join(input_path, "expanded_test_backup.mdp"))

        os.remove(os.path.join(input_path, "expanded_test.mdp"))
        os.remove(os.path.join(input_path, "expanded_test_backup.mdp"))

    def test_print_params(self, capfd, params_dict):
        # capfd is a fixture in pytest for testing STDOUT
        REXEE = get_REXEE_instance(params_dict)
        REXEE.print_params()
        out_1, err = capfd.readouterr()
        L = ""
        L += "Important parameters of EXEE\n============================\n"
        L += f"Python version: {sys.version}\n"
        L += f"GROMACS executable: {REXEE.gmx_path}\n"  # Easier to pass CI. This is easy to catch anyway
        L += f"GROMACS version: {REXEE.gmx_version}\n"  # Easier to pass CI. This is easy to catch anyway
        L += f"ensemble_md version: {ensemble_md.__version__}\n"
        L += "Simulation inputs: ensemble_md/tests/data/sys.gro, ensemble_md/tests/data/sys.top, ensemble_md/tests/data/expanded.mdp\n"  # noqa: E501
        L += "Verbose log file: True\n"
        L += "Proposal scheme: exhaustive\n"
        L += "Whether to perform weight combination: False\n"
        L += "Type of means for weight combination: simple\n"
        L += "Whether to perform histogram correction: False\n"
        L += "Histogram cutoff for weight correction: 1000\n"
        L += "Number of replicas: 4\nNumber of iterations: 10\n"
        L += "Length of each replica: 1.0 ps\nFrequency for checkpointing: 100 iterations\n"
        L += "Total number of states: 9\n"
        L += "Additionally defined swappable states: None\n"
        L += "Additional grompp arguments: None\n"
        L += "Additional runtime arguments: None\n"
        L += "MDP parameters differing across replicas: None\n"
        L += "Alchemical ranges of each replica in REXEE:\n  - Replica 0: States [0, 1, 2, 3, 4, 5]\n"
        L += "  - Replica 1: States [1, 2, 3, 4, 5, 6]\n  - Replica 2: States [2, 3, 4, 5, 6, 7]\n"
        L += "  - Replica 3: States [3, 4, 5, 6, 7, 8]\n"
        assert out_1 == L

        REXEE.reformatted_mdp = True  # Just to test the case where REXEE.reformatted_mdp is True
        REXEE.print_params(params_analysis=True)
        out_2, err = capfd.readouterr()
        L += "\nWhether to build Markov state models and perform relevant analysis: False\n"
        L += "Whether to perform free energy calculations: False\n"
        L += "The step to used in subsampling the DHDL data in free energy calculations, if any: 1\n"
        L += "The chosen free energy estimator for free energy calculations, if any: MBAR\n"
        L += "The method for estimating the uncertainty of free energies in free energy calculations, if any: propagate\n"  # noqa: E501
        L += "The number of bootstrap iterations in the boostrapping method, if used: 50\n"
        L += "The random seed to use in bootstrapping, if used: None\n"
        L += "Note that the input MDP file has been reformatted by replacing hypens with underscores. The original mdp file has been renamed as *backup.mdp.\n"  # noqa: E501
        assert out_2 == L

    def test_initialize_MDP(self, params_dict):
        REXEE = get_REXEE_instance(params_dict)
        MDP = REXEE.initialize_MDP(2)  # the third replica
        assert MDP["nsteps"] == 500
        assert all(
            [
                a == b
                for a, b in zip(
                    MDP["vdw_lambdas"], [0.00, 0.00, 0.00, 0.25, 0.50, 0.75]
                )
            ]
        )
        assert all(
            [
                a == b
                for a, b in zip(
                    MDP["coul_lambdas"], [0.50, 0.75, 1.00, 1.00, 1.00, 1.00]
                )
            ]
        )
        assert all(
            [a == b for a, b in zip(MDP["init_lambda_weights"], [0, 0, 0, 0, 0, 0])]
        )

    def test_update_MDP(self, params_dict):
        new_template = "ensemble_md/tests/data/expanded.mdp"
        iter_idx = 3
        states = [2, 5, 7, 4]
        wl_delta = [0.4, 0.32, 0.256, 0.32]
        weights = [
            [0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0],
            [3.48, 2.78, 3.21, 4.56, 8.79, 0.48],
            [8.45, 0.52, 3.69, 2.43, 4.56, 6.73], ]

        REXEE = get_REXEE_instance(params_dict)
        REXEE.equil = [-1, 1, 0, -1]  # i.e. the 3rd replica will use fixed weights in the next iteration
        MDP_1 = REXEE.update_MDP(
            new_template, 2, iter_idx, states, wl_delta, weights)  # third replica
        MDP_2 = REXEE.update_MDP(
            new_template, 3, iter_idx, states, wl_delta, weights)  # fourth replica

        assert MDP_1["tinit"] == MDP_2["tinit"] == 3
        assert MDP_1["nsteps"] == MDP_2["nsteps"] == 500
        assert MDP_1["init_lambda_state"] == 5
        assert MDP_2["init_lambda_state"] == 1
        assert (MDP_1["init_wl_delta"] == MDP_1["wl_scale"] == MDP_1["wl_ratio"] == "")
        assert (MDP_1["lmc_weights_equil"] == MDP_1["weight_equil_wl_delta"] == "")
        assert MDP_2["init_wl_delta"] == 0.32
        assert all(
            [
                a == b
                for a, b in zip(
                    MDP_1["init_lambda_weights"], [3.48, 2.78, 3.21, 4.56, 8.79, 0.48]
                )
            ]
        )
        assert all(
            [
                a == b
                for a, b in zip(
                    MDP_2["init_lambda_weights"], [8.45, 0.52, 3.69, 2.43, 4.56, 6.73]
                )
            ]
        )

    def test_extract_final_dhdl_info(self, params_dict):
        REXEE = get_REXEE_instance(params_dict)
        dhdl_files = [
            os.path.join(input_path, f"dhdl/dhdl_{i}.xvg") for i in range(REXEE.n_sim)
        ]
        states = REXEE.extract_final_dhdl_info(dhdl_files)
        assert states == [5, 2, 2, 8]

    def test_extract_final_log_info(self, params_dict):
        REXEE = get_REXEE_instance(params_dict)
        log_files = [
            os.path.join(input_path, f"log/EXE_{i}.log") for i in range(REXEE.n_sim)]
        wl_delta, weights, counts = REXEE.extract_final_log_info(log_files)
        assert wl_delta == [0.4, 0.5, 0.5, 0.5]
        assert np.allclose(weights, [
            [0, 1.03101, 2.55736, 3.63808, 4.47220, 6.13408],
            [0, 1.22635, 2.30707, 2.44120, 4.10308, 6.03106],
            [0, 0.66431, 1.25475, 0.24443, 0.59472, 0.70726],
            [0, 0.09620, 1.59937, -4.31679, -22.89436, -28.08701], ])
        assert counts == [
            [4, 11, 9, 9, 11, 6],
            [9, 8, 8, 11, 7, 7],
            [3, 1, 1, 9, 15, 21],
            [0, 0, 0, 1, 18, 31], ]
        assert REXEE.equil == [-1, -1, -1, -1]

    def test_get_averaged_weights(self, params_dict):
        REXEE = get_REXEE_instance(params_dict)
        log_files = [
            os.path.join(input_path, f"log/EXE_{i}.log") for i in range(REXEE.n_sim)]
        avg, err = REXEE.get_averaged_weights(log_files)
        assert np.allclose(avg[0],  [0, 2.55101, 3.35736, 4.83808, 4.8722, 5.89408])
        assert np.allclose(err[0], [0, 1.14542569, 1.0198039, 0.8, 0.69282032, 0.35777088])

    def test_identify_swappable_pairs(self, params_dict):
        REXEE = get_REXEE_instance(params_dict)
        REXEE.state_ranges = [list(range(i, i + 5)) for i in range(REXEE.n_sim)]  # 5 states per replica
        states = [4, 2, 2, 7]   # This would lead to the swappables: [(0, 1), (0, 2), (1, 2)] in the standard case

        # Case 1: Any case that is not neighboring swap has the same definition for the swappable pairs
        swappables_1 = REXEE.identify_swappable_pairs(states, REXEE.state_ranges, REXEE.proposal == 'neighboring')
        assert swappables_1 == [(0, 1), (0, 2), (1, 2)]

        # Case 2: Neighboring exchange
        REXEE.proposal = 'neighboring'
        swappables_2 = REXEE.identify_swappable_pairs(states, REXEE.state_ranges, REXEE.proposal == 'neighboring')
        assert swappables_2 == [(0, 1), (1, 2)]

    def test_propose_swap(self, params_dict):
        random.seed(0)
        REXEE = get_REXEE_instance(params_dict)
        swap_1 = REXEE.propose_swap([])
        swap_2 = REXEE.propose_swap([(0, 1), (0, 2), (1, 2)])
        assert swap_1 == []
        assert swap_2 == (1, 2)

    def test_get_swapping_pattern(self, params_dict):
        # weights are obtained from the log files in data/log, where the last states are 5, 2, 2, 8 (global indices)
        # state_ranges are: 0-5, 1-6, ..., 3-8
        dhdl_files = [os.path.join(input_path, f"dhdl/dhdl_{i}.xvg") for i in range(4)]

        # Case 1: Empty swap list
        REXEE = get_REXEE_instance(params_dict)
        REXEE.verbose = False
        states = [0, 6, 7, 8]  # No swappable pairs
        f = copy.deepcopy(dhdl_files)
        pattern_1, swap_list_1 = REXEE.get_swapping_pattern(f, states)
        assert REXEE.n_empty_swappable == 1
        assert REXEE.n_swap_attempts == 0
        assert REXEE.n_rejected == 0
        assert pattern_1 == [0, 1, 2, 3]
        assert swap_list_1 == []

        # Case 2: Single swap (proposal = 'single')
        random.seed(0)
        REXEE = get_REXEE_instance(params_dict)
        REXEE.verbose = True
        REXEE.proposal = 'single'  # n_ex will be set to 1 automatically.
        states = [5, 2, 2, 8]  # swappable pairs: [(0, 1), (0, 2), (1, 2)], swap = (1, 2), accept
        f = copy.deepcopy(dhdl_files)
        pattern_2, swap_list_2 = REXEE.get_swapping_pattern(f, states)
        assert REXEE.n_swap_attempts == 1
        assert REXEE.n_rejected == 0
        assert pattern_2 == [0, 2, 1, 3]
        assert swap_list_2 == [(1, 2)]

        # Case 3: Neighboring swap
        random.seed(0)
        REXEE = get_REXEE_instance(params_dict)
        REXEE.proposal = 'neighboring'  # n_ex will be set to 1 automatically.
        states = [5, 2, 2, 8]  # swappable pairs: [(0, 1), (0, 2), (1, 2)], swap = (1, 2), accept
        f = copy.deepcopy(dhdl_files)
        pattern_3, swap_list_3 = REXEE.get_swapping_pattern(f, states)
        assert REXEE.n_swap_attempts == 1
        assert REXEE.n_rejected == 0
        assert pattern_3 == [0, 2, 1, 3]
        assert swap_list_3 == [(1, 2)]

        # Case 4-1: Exhaustive swaps that end up in a single swap
        random.seed(0)
        REXEE = get_REXEE_instance(params_dict)
        REXEE.proposal = 'exhaustive'
        states = [5, 2, 2, 8]  # swappable pairs: [(0, 1), (0, 2), (1, 2)], swap = (1, 2), accept
        f = copy.deepcopy(dhdl_files)
        pattern_4_1, swap_list_4_1 = REXEE.get_swapping_pattern(f, states)
        assert REXEE.n_swap_attempts == 1
        assert REXEE.n_rejected == 0
        assert pattern_4_1 == [0, 2, 1, 3]
        assert swap_list_4_1 == [(1, 2)]

        # Case 4-2: Exhaustive swaps that involve multiple attempted swaps
        random.seed(0)
        REXEE = get_REXEE_instance(params_dict)
        REXEE.proposal = 'exhaustive'
        states = [4, 2, 4, 3]  # swappable pairs: [(0, 1), (0, 2), (0, 3), (1, 2), (2, 3)]; swap 1: (2, 3), accepted; swap 2: (0, 1), accept  # noqa: E501
        f = copy.deepcopy(dhdl_files)
        pattern_4_2, swap_list_4_2 = REXEE.get_swapping_pattern(f, states)
        assert REXEE.n_swap_attempts == 2   # \Delta is negative for both swaps -> both accepted
        assert REXEE.n_rejected == 0
        assert pattern_4_2 == [1, 0, 3, 2]
        assert swap_list_4_2 == [(2, 3), (0, 1)]

    def test_calc_prob_acc(self, capfd, params_dict):
        # k = 1.380649e-23; NA = 6.0221408e23; T = 298; kT = k * NA * T / 1000 = 2.4777098766670016
        REXEE = get_REXEE_instance(params_dict)
        # REXEE.state_ranges = [[0, 1, 2, 3, 4, 5], [1, 2, 3, 4, 5, 6], ..., [3, 4, 5, 6, 7, 8]]
        states = [5, 2, 2, 8]
        shifts = [0, 1, 2, 3]
        dhdl_files = [os.path.join(input_path, f"dhdl/dhdl_{i}.xvg") for i in range(4)]

        # Test 1
        swap = (0, 1)
        prob_acc_1 = REXEE.calc_prob_acc(swap, dhdl_files, states, shifts)
        out, err = capfd.readouterr()
        # dU = (-9.1366697  + 11.0623788)/2.4777098766670016 ~ 0.7772 kT, so p_acc = 0.45968522728859024
        assert prob_acc_1 == pytest.approx(0.45968522728859024)
        assert 'U^i_n - U^i_m = -3.69 kT, U^j_m - U^j_n = 4.46 kT, Total dU: 0.78 kT' in out

        # Test 2
        swap = (0, 2)
        prob_acc_2 = REXEE.calc_prob_acc(swap, dhdl_files, states, shifts)
        out, err = capfd.readouterr()
        # dU = (-9.1366697 + 4.9963939)/2.4777098766670016 ~ -1.6710 kT, so p_acc = 1
        assert prob_acc_2 == 1

    def test_accept_or_reject(self, params_dict):
        REXEE = get_REXEE_instance(params_dict)
        random.seed(0)
        swap_bool_1 = REXEE.accept_or_reject(0)
        swap_bool_2 = REXEE.accept_or_reject(0.8)  # rand = 0.844
        swap_bool_3 = REXEE.accept_or_reject(0.8)  # rand = 0.758

        assert REXEE.n_swap_attempts == 0  # since we didn't use get_swapping_pattern
        assert REXEE.n_rejected == 2
        assert swap_bool_1 is False
        assert swap_bool_2 is False
        assert swap_bool_3 is True

    def test_weight_correction(self, params_dict):
        REXEE = get_REXEE_instance(params_dict)

        # Case 1: Perform weight correction (N_cutoff reached)
        REXEE.N_cutoff = 5000
        REXEE.verbose = False  # just to increase code coverage
        weights_1 = [[0, 10.304, 20.073, 29.364]]
        counts_1 = [[31415, 45701, 55457, 59557]]
        weights_1 = REXEE.weight_correction(weights_1, counts_1)
        assert np.allclose(weights_1, [
            [
                0,
                10.304 + np.log(31415 / 45701),
                20.073 + np.log(45701 / 55457),
                29.364 + np.log(55457 / 59557),
            ]
        ])  # noqa: E501

        # Case 2: Perform weight correction (N_cutoff not reached by both N_k and N_{k-1})
        REXEE.verbose = True
        weights_2 = [[0, 10.304, 20.073, 29.364]]
        counts_2 = [[3141, 4570, 5545, 5955]]
        weights_2 = REXEE.weight_correction(weights_2, counts_2)
        assert np.allclose(weights_2, [[0, 10.304, 20.073, 29.364 + np.log(5545 / 5955)]])

    def test_combine_weights(self, params_dict):
        """
        Here we just test the combined weights.
        """
        REXEE = get_REXEE_instance(params_dict)
        REXEE.n_tot = 6
        REXEE.n_sub = 4
        REXEE.s = 1
        REXEE.n_sim = 3
        REXEE.state_ranges = [[0, 1, 2, 3], [1, 2, 3, 4], [2, 3, 4, 5]]

        # Test 1: simple means
        weights = [[0, 2.1, 4.0, 3.7], [0, 1.7, 1.2, 2.6], [0, -0.4, 0.9, 1.9]]
        w_1, g_vec_1 = REXEE.combine_weights(weights)
        assert np.allclose(w_1, [
            [0, 2.1, 3.9, 3.5],
            [0, 1.8, 1.4, 2.75],
            [0, -0.4, 0.95, 1.95]])
        assert np.allclose(list(g_vec_1), [0, 2.1, 3.9, 3.5, 4.85, 5.85])

        # Test 2: weighted means
        weights = [[0, 2.1, 4.0, 3.7], [0, 1.7, 1.2, 2.6], [0, -0.4, 0.9, 1.9]]
        errors = [[0, 0.1, 0.15, 0.1], [0, 0.12, 0.1, 0.12], [0, 0.12, 0.15, 0.1]]
        w_2, g_vec_2 = REXEE.combine_weights(weights, errors)
        assert np.allclose(w_2, [
            [0, 2.1, 3.86140725, 3.45417313],
            [0, 1.76140725, 1.35417313, 2.71436889],
            [0, -0.40723412, 0.95296164, 1.95296164]])
        assert np.allclose(list(g_vec_2), [0, 2.1, 3.861407249466951, 3.4541731330165306, 4.814368891580968, 5.814368891580968])  # noqa: E501

    def test_histogram_correction(self, params_dict):
        REXEE = get_REXEE_instance(params_dict)
        REXEE.n_tot = 6
        REXEE.n_sub = 5
        REXEE.s = 1
        REXEE.n_sim = 2
        REXEE.state_ranges = [[0, 1, 2, 3, 4], [1, 2, 3, 4, 5]]
        hist = [[416, 332, 130, 71, 61], [303, 181, 123, 143, 260]]

        hist_modified = REXEE.histogram_correction(hist)
        assert hist_modified == [[416, 332, 161, 98, 98], [332, 161, 98, 98, 178]]



####################################################################
#                                                                  #
#    ensemble_md,                                                  #
#    a python package for running GROMACS simulation ensembles     #
#                                                                  #
#    Written by Wei-Tse Hsu <wehs7661@colorado.edu>                #
#    Copyright (c) 2022 University of Colorado Boulder             #
#                                                                  #
####################################################################
import numpy as np
from ensemble_md.utils.utils import run_gmx_cmd


def cluster_traj(gmx_executable, inputs, grps, coupled_only=True, method='linkage', cutoff=0.1, suffix=None):
    """
    Performs clustering analysis on a trajectory using the GROMACS command :code:`gmx cluster`.
    Note that only fully coupled configurations are considered.

    Parameters
    ----------
    gmx_executable : str
        The path to the GROMACS executable.
    inputs : dict
        A dictionary that contains the different input files required for the clustering analysis.
        The dictionary must have the following four keys: :code:`traj` (input trajectory file in
        XTC or TRR format), :code:`config` (the configuration file in TPR or GRO format),
        :code:`xvg` (a GROMACS XVG file), and :code:`index` (an index/NDX file), with the values
        being the paths. Note that the value of the key :code:`index` can be :code:`None`, in which
        case the function will use a default index file generated by :code:`gmx make_ndx`. If the
        parameter :code:`coupled_only` is set to :code:`True`, an XVG file that contains the time
        series of the state index (e.g., :code:`dhdl.xvg`) must be provided with the key :code:`xvg`.
        Otherwise, the key :code:`xvg` can be set to :code:`None`.
    grps : dict
        A dictionary that contains the names of the groups in the index file (NDX) for
        centering the system, calculating the RMSD, and outputting. The keys are
        :code:`center`, :code:`rmsd`, and :code:`output`.
    coupled_only : bool
        Whether to consider only the fully coupled configurations. The default is :code:`True`.
    method : str
        The method for clustering available for the GROMACS command :code:`gmx cluster`. The default is 'linkage'.
        Check the GROMACS documentation for other available options.
    cutoff : float
        The RMSD cutoff for clustering in nm. The default is 0.1.
    suffix : str
        The suffix for the output files. The default is :code:`None`, which means no suffix will be added.
    """
    # Check if the index file is provided
    if inputs['index'] is None:
        print('Running gmx make_ndx to generate an index file ...')
        args = [
            gmx_executable, 'make_ndx',
            '-f', inputs['config'],
            '-o', 'index.ndx',
        ]
        returncode, stdout, stderr = run_gmx_cmd(args, prompt_input='q\n')
        inputs['index'] = 'index.ndx'

    # Check if the groups are present in the index file
    with open(inputs['index'], 'r') as f:
        content = f.read()
    for key in grps:
        if grps[key] not in content:
            raise ValueError(f'The group {grps[key]} is not present in the provided/generated index file.')

    outputs = {
        'nojump': 'nojump.xtc',
        'center': 'center.xtc',
        'rmsd-clust': 'rmsd-clust.xpm',
        'rmsd-dist': 'rmsd-dist.xvg',
        'cluster-log': 'cluster.log',
        'cluster-pdb': 'clusters.pdb',
        'rmsd': 'rmsd.xvg',  # inter-medoid RMSD
    }
    if suffix is not None:
        for key in outputs:
            outputs[key] = outputs[key].replace('.', f'_{suffix}.')

    # Check if there is any fully coupled state in the trajectory
    lambda_data = np.transpose(np.loadtxt(inputs['xvg'], comments=['#', '@']))[1]
    if coupled_only is True and 0 not in lambda_data:
        print('Terminating clustering analysis since no fully decoupled state is present in the input trajectory while coupled_only is set to True.')  # noqa: E501
    else:
        # Either coupled_only is False or coupled_only is True but there are coupled configurations.
        print('Eliminating jumps across periodic boundaries for the input trajectory ...')
        args = [
            gmx_executable, 'trjconv',
            '-f', inputs['traj'],
            '-s', inputs['config'],
            '-n', inputs['index'],
            '-o', outputs['nojump'],
            '-center', 'yes',
            '-pbc', 'nojump',
        ]

        if coupled_only:
            if inputs['xvg'] is None:
                raise ValueError('The parameter "coupled_only" is set to True but no XVG file is provided.')
            args.extend([
                '-drop', inputs['xvg'],
                '-dropover', '0'
            ])

        returncode, stdout, stderr = run_gmx_cmd(args, prompt_input=f'{grps["center"]}\n{grps["output"]}\n')
        if returncode != 0:
            print(f'Error with return code: {returncode}):\n{stderr}')

        print('Centering the system ...')
        args = [
            gmx_executable, 'trjconv',
            '-f', outputs['nojump'],
            '-s', inputs['config'],
            '-n', inputs['index'],
            '-o', outputs['center'],
            '-center', 'yes',
            '-pbc', 'mol',
            '-ur', 'compact',
        ]
        returncode, stdout, stderr = run_gmx_cmd(args, prompt_input=f'{grps["center"]}\n{grps["output"]}\n')
        if returncode != 0:
            print(f'Error with return code: {returncode}):\n{stderr}')

        if coupled_only is True:
            N_coupled = np.count_nonzero(lambda_data == 0)
            print(f'Number of fully coupled configurations: {N_coupled}')

        print('Performing clustering analysis ...')
        args = [
            gmx_executable, 'cluster',
            '-f', outputs['center'],
            '-s', inputs['config'],
            '-n', inputs['index'],
            '-o', outputs['rmsd-clust'],
            '-dist', outputs['rmsd-dist'],
            '-g', outputs['cluster-log'],
            '-cl', outputs['cluster-pdb'],
            '-cutoff', str(cutoff),
            '-method', method,
        ]
        returncode, stdout, stderr = run_gmx_cmd(args, prompt_input=f'{grps["rmsd"]}\n{grps["output"]}\n')
        if returncode != 0:
            print(f'Error with return code: {returncode}):\n{stderr}')

        rmsd_range, rmsd_avg, n_clusters = get_cluster_info(outputs['cluster-log'])

        print(f'Range of RMSD values: from {rmsd_range[0]:.3f} to {rmsd_range[1]:.3f} nm')
        print(f'Average RMSD: {rmsd_avg:.3f} nm')
        print(f'Number of clusters: {n_clusters}')

        if n_clusters > 1:
            clusters, sizes = get_cluster_members(outputs['cluster-log'])
            for i in range(1, n_clusters + 1):
                print(f'  - Cluster {i} accounts for {sizes[i] * 100:.2f}% of the total configurations.')

            n_transitions, t_transitions = count_transitions(clusters)
            print(f'Number of transitions between the two biggest clusters: {n_transitions}')
            print(f'Time frames of the transitions (ps): {t_transitions}')

            print('Calculating the inter-medoid RMSD between the two biggest clusters ...')
            # Note that we pass outputs['cluster-pdb'] to -s so that the first medoid will be used as the reference
            args = [
                gmx_executable, 'rms',
                '-f', outputs['cluster-pdb'],
                '-s', outputs['cluster-pdb'],
                '-o', outputs['rmsd'],
            ]
            if inputs['index'] is not None:
                args.extend(['-n', inputs['index']])

            # Here we simply assume same groups for least-squares fitting and RMSD calculation
            returncode, stdout, stderr = run_gmx_cmd(args, prompt_input=f'{grps["rmsd"]}\n{grps["rmsd"]}\n')
            if returncode != 0:
                print(f'Error with return code: {returncode}):\n{stderr}')

            rmsd = np.transpose(np.loadtxt(outputs['rmsd'], comments=['@', '#']))[1][1]  # inter-medoid RMSD
            print(f'Inter-medoid RMSD between the two biggest clusters: {rmsd:.3f} nm')


def get_cluster_info(cluster_log):
    """
    Gets the metadata of the LOG file generated by the GROMACS :code:`gmx cluster` command.

    Parameters
    ----------
    cluster_log : str
        The LOG file generated by the GROMACS :code:`gmx cluster` command.

    Returns
    -------
    rmsd_range: list
        The range of RMSD values
    rmsd_avg: float
        The average RMSD value.
    n_clusters : int
        The number of clusters.
    """
    f = open(cluster_log, 'r')
    lines = f.readlines()
    f.close()

    rmsd_range = []
    for line in lines:
        if 'The RMSD ranges from' in line:
            rmsd_range.append(float(line.split('from')[-1].split('to')[0]))
            rmsd_range.append(float(line.split('from')[-1].split('to')[-1].split('nm')[0]))
        if 'Average RMSD' in line:
            rmsd_avg = float(line.split('is')[-1])
        if 'Found' in line:
            n_clusters = int(line.split()[1])
            break

    return rmsd_range, rmsd_avg, n_clusters


def get_cluster_members(cluster_log):
    """
    Gets the members of each cluster from the LOG file generated by the GROMACS :code:`gmx cluster` command.

    Parameters
    ----------
    cluster_log : str
        The LOG file generated by the GROMACS :code:`gmx cluster` command.

    Returns
    -------
    clusters : dict
        A dictionary that contains the cluster index (starting from 1) as the key and the list of members
        (configurations at different timeframes) as the value.
    sizes : dict
        A dictionary that contains the cluster index (starting from 1) as the key and the size of the cluster
        (in fraction) as the value.
    """
    clusters = {}
    current_cluster = 0
    start_processing = False

    f = open(cluster_log, 'r')
    lines = f.readlines()
    f.close()

    for line in lines:
        # Start processing when we reach the line that starts with "cl."
        if line.strip().startswith("cl."):
            start_processing = True
            continue  # Skip this line and continue to the next iteration

        if start_processing:
            parts = line.split('|')
            try:
                current_cluster = int(parts[0].strip())
                clusters[current_cluster] = []
            except ValueError:
                pass

            # This is either a new cluster or continuation of it, add members
            members = parts[-1].split()
            clusters[current_cluster].extend([int(i) for i in members])

    sizes_list = [len(clusters[i]) for i in clusters]
    sizes = {i: sizes_list[i - 1] / sum(sizes_list) for i in clusters}

    return clusters, sizes


def count_transitions(clusters):
    """
    Counts the number of transitions between the two biggest clusters.

    Parameters
    ----------
    clusters : dict
        A dictionary that contains the cluster index (starting from 1) as the key and the list of members
        (configurations at different timeframes) as the value.

    Returns
    -------
    n_transitions : int
        The number of transitions between the two biggest clusters.
    t_transitions : list
        The list of time frames when the transitions occur.
    """
    # Combine and sort all cluster members for the first two biggest clusters while keeping track of their origin
    all_members = [(member, 1) for member in clusters[1]] + [(member, 2) for member in clusters[2]]
    all_members.sort()

    # Count transitions and record time frames
    n_transitions = 0
    t_transitions = []
    last_cluster = all_members[0][1]  # the cluster index of the last time frame in the previous iteration
    for member in all_members[1:]:
        if member[1] != last_cluster:
            n_transitions += 1
            last_cluster = member[1]
            t_transitions.append(member[0])

    return n_transitions, t_transitions



####################################################################
#                                                                  #
#    ensemble_md,                                                  #
#    a python package for running GROMACS simulation ensembles     #
#                                                                  #
#    Written by Wei-Tse Hsu <wehs7661@colorado.edu>                #
#    Copyright (c) 2022 University of Colorado Boulder             #
#                                                                  #
####################################################################
import sys
import random
import argparse
import numpy as np
import pandas as pd
from ensemble_md.replica_exchange_EE import ReplicaExchangeEE


def initialize(args):
    parser = argparse.ArgumentParser(
        description='This code explores the parameter space of homogenous REXEE to help you figure \
                out all possible combinations of the number of replicas, the number of \
                states in each replica, and the number of overlapping states, and the total number states.')
    parser.add_argument('-N',
                        '--N',
                        required=True,
                        type=int,
                        help='The total number of states of the REXEE simulation.')
    parser.add_argument('-r',
                        '--r',
                        type=int,
                        help='The number of replicas that compose the REXEE simulation.')
    parser.add_argument('-n',
                        '--n',
                        type=int,
                        help='The number of states for each replica.')
    parser.add_argument('-s',
                        '--s',
                        type=int,
                        help='The state shift between adjacent replicas.')
    parser.add_argument('-c',
                        '--cnst',
                        default=False,
                        action='store_true',
                        help='Whether the apply the constraint such that the number of overlapping states \
                            does not exceed 50%% of the number of states in both overlapping replicas.')
    parser.add_argument('-e',
                        '--estimate',
                        default=False,
                        action='store_true',
                        help='Whether to provide estimates of the chance of not having any swappable \
                            pairs for each solution.')
    args_parse = parser.parse_args(args)

    return args_parse


def solv_REXEE_diophantine(N, constraint=False):
    """
    Solves the general nonlinear Diophantine equation associated with the homogeneous REXEE
    parameters. Specifically, given the total number of states :math:`N` and the number of replicas
    r, the states for each replica n and the state shift s can be expressed as:
    n = N + (r-1)(t-1), and s = 1 - t, with the range of t being either the following:
    - Without the additional constraint, (r-N+1)/r <= t <= 0
    - With the additional constraint, (r-N+1)/r <= t <= (r-N+1)/(r+1)

    Parameters
    ----------
    N : int
        The total number of states of the homogeneous REXEE of interesst.
    constraint : bool
        Whether to apply additional constraints such that n-s <= 1/2n.

    Returns
    -------
    soln_all : pd.DataFrame
        A pandas DataFrame that lists all the solutions of (N, r, n, s).
    """
    soln_all = []   # [N, r, n, s]
    r_list = range(2, N)
    for r in r_list:
        if constraint is False:
            # To include the upper bound 0, we set `stop=1` instead of 0.
            t = np.arange(int((r - N + 1) / r), 1).astype(int)
        else:
            if (r - N + 1) / (r + 1) == int((r - N + 1) / (r + 1)):
                # If the upper bound is a negative integer, we need to shift `stop` by 1.
                t = np.arange(int((r - N + 1) / r), int((r - N + 1) / (r + 1)) + 1).astype(int)
            else:
                # If the upper bound is a negative floating number, there is no need to shift.
                t = np.arange(int((r - N + 1) / r), int((r - N + 1) / (r + 1))).astype(int)
        n = N + (r - 1) * (t - 1)
        s = 1 - t
        soln_all.extend([{'N': N, 'r': r, 'n': n[i], 's': s[i]} for i in range(len(t))])  # [N, r, n, s]
    soln_all = pd.DataFrame.from_dict(soln_all)  # turn the dict to pandas df, which is more convenient

    return soln_all


def estimate_swapless_rate(state_ranges, N=1000000):
    """
    Estimates the chance of not having any swappable pair in an iteration.

    Parameters
    ----------
    state_ranges : list
        A list of lists of state indices. (Like the attribute :code:`state_ranges` in :code:`EnsemblEXE`.)
    N : n
        The number of Monte Carlo iterations for the estimation.

    Returns
    -------
    p : float
        The probability of not having any swappable pair in an iteration.
    """
    n = 0  # number of times of not having any swappable pairs
    for i in range(N):
        rands = [random.choice(state_ranges[i]) for i in range(len(state_ranges))]
        swappables = ReplicaExchangeEE.identify_swappable_pairs(rands, state_ranges, False)
        if swappables == []:
            n += 1

    p = n / N * 100   # in percentage

    return p


def main():
    # For now, we only consider homogenous REXEE simulations
    args = initialize(sys.argv[1:])
    print('Exploration of the REXEE parameter space')
    print('=======================================')
    print('[ REXEE parameters of interest ]')
    print('- N: The total number of states')
    print('- r: The number of replicas')
    print('- n: The number of states for each replica')
    print('- s: The state shift between adjacent replicas')

    # Enuerate all possible combinations of (N, r, n, s) even if any of r, n, s is given - it's easy/fast anyway.
    soln_all = solv_REXEE_diophantine(args.N, constraint=args.cnst)

    # Now filter the solutions
    if args.r is not None:
        soln_all = soln_all[soln_all['r'] == args.r]
    if args.n is not None:
        soln_all = soln_all[soln_all['n'] == args.n]
    if args.s is not None:
        soln_all = soln_all[soln_all['s'] == args.s]

    # Print out the solutions
    print('\n[ Solutions ]')
    for row_idx in range(len(soln_all)):
        soln = soln_all.iloc[row_idx]
        start_idx = [i * soln['s'] for i in range(soln['r'])]
        state_ranges = [list(np.arange(i, i + soln['n'])) for i in start_idx]

        if args.estimate is True:
            p = estimate_swapless_rate(state_ranges)
            print(f"- Solution {row_idx + 1}: (N, r, n, s) = ({args.N}, {soln['r']}, {soln['n']}, {soln['s']}), {p:.1f}% chance to not have any swappable pair")  # noqa: E501
        else:
            print(f"- Solution {row_idx + 1}: (N, r, n, s) = ({args.N}, {soln['r']}, {soln['n']}, {soln['s']})")
        for i in range(soln['r']):
            print(f'  - Replica {i}: {state_ranges[i]}')
        print()
